from setuptools import setup, find_packages

setup(
    name='speac',
    version='1.0.10',
    packages=find_packages(),
    url='https://github.com/GolzitskyNikolay/SPEAC-analysis/',
    license='LICENSE.txt',
    long_discription="Readme.md",
    author='GlazNik',
    author_email='golzitskij.ns@edu.spbstu.ru',
    description='David Cope\'s SPEAC-analysis in Python',
    python_requires='>=3.0'
)
